import functools
import logging
import re

# Parser


class Token:
    re_element_tag = re.compile(r"^E\:\s([^\s]+)\s\(([^\)]+)\)")
    re_attribute = re.compile(r"^A:\s([^=]*)=([^$]+)")
    re_parentheses = re.compile(r"\([^)]*\)")

    def __init__(self, text):
        self._stripped = text.strip()
        self._bad = False

        if len(text) == 0:
            self._bad = True
        else:
            try:
                self._deep = len(text) - len(text.lstrip())
                self._type = self._stripped[0]
            except Exception as e:
                logging.warning(e)
                self._bad = True

    def as_element(self):
        g = Token.re_element_tag.match(self._stripped).groups()
        return g

    def as_attribute(self):
        g = Token.re_attribute.match(Token.remove_parentheses(self._stripped)).groups()
        return g

    @staticmethod
    def remove_parentheses(s):
        return Token.re_parentheses.sub("", s)

    def __str__(self):
        return "{}, {}, {}".format(self._deep, self._type, self._stripped)

    @property
    def type(self):
        return self._type

    @property
    def deep(self):
        return self._deep

    @property
    def raw(self):
        return self._stripped

    @property
    def bad(self):
        return self._bad


class ManifestAttribute:
    def __init__(self, element, token):
        self._element = element

        r = token.as_attribute()

        self._name = r[0].strip()
        self._value = r[1].strip()
        self._origin = token.raw

    def __str__(self):
        return "{}={}".format(self._name, self._value)

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value


class ManifestNamespace:
    def __init__(self, element, token):
        pass


class ManifestElement:
    def __init__(self, parent=None, token=None):

        self._namespaces: [ManifestNamespace] = []
        self._attributes: dict[str, ManifestAttribute] = {}
        self._elements: [ManifestElement] = []

        self._parent = parent

        if token is not None:
            e = token.as_element()
            self._name = e[0]
            self._at_line = e[1]
            self._deep = token.deep
            self._type = self._name
        else:
            self._name = '[ROOT]'
            self._type = '[ROOT]'
            self._deep = -1

    def __str__(self):
        intent = self._deep
        if intent < 0:
            intent = 0

        intent_str = " " * intent

        s = "{}<{}>".format(intent_str, self._name)

        for a in self._attributes:
            s += "\n{}[{}={}]".format(intent_str, str(a), str(self._attributes[a].value))

        for e in self._elements:
            s += "\n{}".format(str(e))

        return s

    def add_element(self, element):
        self._elements.append(element)

    def get_elements(self):
        return self._elements

    def add_attribute(self, attribute: ManifestAttribute):
        self._attributes[attribute.name] = attribute

    def get_attribute_value(self, key, default):
        r = self._attributes.get(key)
        if r is not None:
            return r.value
        else:
            return default

    def add_namespace(self, namespace):
        self._namespaces.append(namespace)

    @property
    def name(self):
        return self._name

    @property
    def deep(self):
        return self._deep


class Manifest(ManifestElement):

    def __init__(self, text):
        super().__init__()
        lines = text.split('\n')

        stack: [ManifestElement] = [self]

        for line in lines:
            token = Token(line)

            if token.bad:
                logging.debug("bad token {}".format(token.raw))
                continue

            if token.type == 'N':
                stack[-1].add_namespace(ManifestNamespace(stack[-1], token))
            elif token.type == 'E':
                element = ManifestElement(stack[-1], token)
                while stack[-1].deep >= element.deep:
                    stack.pop()
                stack[-1].add_element(element)
                stack.append(element)
            elif token.type == 'A':
                stack[-1].add_attribute(ManifestAttribute(stack[-1], token))
            else:
                logging.warning("unknown tag '{}'".format(token.raw))


# Wrappers
class Attribute:
    def __init__(self, name, default):
        self._name = name
        self._default = default

    def __call__(self, fn):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            data: ManifestElement = args[0].data
            return fn(args[0], data.get_attribute_value(self._name, self._default))

        return decorated


class Wrapper:
    def __init__(self, data: ManifestElement):
        self._data = data

    @property
    def data(self):
        return self._data

    @property
    @Attribute("android:name", "Unknown")
    def name(self, value):
        return value


class ManifestWrapper(Wrapper):

    def __init__(self, manifest: ManifestElement):

        for e in manifest.get_elements():
            if e.name == 'manifest':
                self._manifest = e

        if self._manifest is None:
            logging.error("no manifest tag")
            return

        super().__init__(self._manifest)

        self._use_features = []
        self._application = None

        for e in self.data.get_elements():
            if e.name == 'use_feature':
                self._use_features.append(UseFeature(e))
            elif e.name == 'application':
                if self._application is not None:
                    logging.warning("multi application in manifest")
                self._application = Application(e)

    @property
    def use_features(self):
        return self._use_features

    @property
    def application(self):
        return self._application


class UseFeature(Wrapper):
    def __init__(self, data: ManifestElement):
        super().__init__(data)


class Application(Wrapper):
    def __init__(self, data: ManifestElement):
        super().__init__(data)
        self._meta = []
        self._activities = []
        self._receivers = []
        self._services = []
        self._providers = []
        self._launchActivity = None

        for e in data.get_elements():
            if e.name == 'meta-data':
                self._meta.append(Meta(e))
            elif e.name == 'activity':
                activity = Activity(e)
                self._activities.append(activity)
                if activity.is_launch_activity:
                    if self._launchActivity is not None:
                        logging.warning("multi launchActivity {}", activity.name)
                    self._launchActivity = activity
            elif e.name == 'receiver':
                self._receivers.append(Receiver(e))
            elif e.name == 'service':
                self._services.append(Service(e))
            elif e.name == 'provider':
                self._providers.append(Provider(e))

    @property
    @Attribute("android:debuggable", 0x0)
    def debuggable(self, value):
        return not int(value, 16) == 0

    @property
    @Attribute("android:largeHeap", 0x0)
    def largeHeap(self, value):
        return not int(value, 16) == 0

    @property
    def activities(self):
        return self._activities

    @property
    def services(self):
        return self._services

    @property
    def metas(self):
        return self._meta

    @property
    def receivers(self):
        return self._receivers

    @property
    def providers(self):
        return self._providers


class Activity(Wrapper):
    def __init__(self, data: ManifestElement):
        super().__init__(data)

        self._intent_filter: IntentFilter = None
        self._meta: [Meta] = []

        for e in self.data.get_elements():
            if e.name == 'intent-filter':
                if self._intent_filter is not None:
                    logging.warning("multi intent-filter, activity: {}".format(self.name))
                self._intent_filter = IntentFilter(e)
            elif e.name == 'meta-data':
                self._meta.append(Meta(e))

    @property
    def is_launch_activity(self):
        return self._intent_filter is not None and self._intent_filter.is_launch

    @property
    @Attribute("android:name", "Unknown")
    def name(self, value):
        return value[1:-1]


class Service(Wrapper):
    def __init__(self, data: ManifestElement):
        super().__init__(data)

    @property
    @Attribute("android:name", "Unknown")
    def name(self, value):
        return value[1:-1]


class Receiver(Wrapper):
    def __init__(self, data: ManifestElement):
        super().__init__(data)

    @property
    @Attribute("android:name", "Unknown")
    def name(self, value):
        return value[1:-1]


class Provider(Wrapper):
    def __init__(self, data: ManifestElement):
        super().__init__(data)

    @property
    @Attribute("android:name", "Unknown")
    def name(self, value):
        return value[1:-1]


class Meta(Wrapper):
    def __init__(self, data: ManifestElement):
        super().__init__(data)

    @property
    @Attribute("android:name", None)
    def name(self, value):
        return value[1:-1]

    @property
    @Attribute("android:value", None)
    def value(self, value):
        return value


class IntentFilter(Wrapper):
    def __init__(self, data: ManifestElement):
        super().__init__(data)

        self._action: Action = None
        self._categories: [Category] = []

        for e in self.data.get_elements():
            if e.name == 'action':
                if self._action is not None:
                    logging.warning("multi action, activity: {}".format(self.name))
                self._action = Action(e)
            elif e.name == 'category':
                self._categories.append(Category(e))

    @property
    def is_launch(self):

        has_launch_cate = False

        for c in self._categories:
            if c.name == Category.LAUNCHER:
                has_launch_cate = True
                break

        return self._action.name == Action.MAIN and has_launch_cate


class Action(Wrapper):
    MAIN = "android.intent.action.MAIN"

    def __init__(self, data: ManifestElement):
        super().__init__(data)

    @property
    @Attribute("android:name", None)
    def name(self, value):
        return value

    @property
    @Attribute("android:name", "Unknown")
    def name(self, value):
        return value[1:-1]


class Category(Wrapper):
    LAUNCHER = "android.intent.category.LAUNCHER"

    def __init__(self, data: ManifestElement):
        super().__init__(data)

    @property
    @Attribute("android:name", None)
    def name(self, value):
        return value

    @property
    @Attribute("android:name", "Unknown")
    def name(self, value):
        return value[1:-1]



import os
import platform
import sys
import subprocess
from pkg_resources import resource_filename

import whichcraft


def get_adb_exe():
    # 1. find in $PATH
    exe = whichcraft.which("adb")
    if exe and _is_valid_exe(exe):
        return exe

    plt = platform.system()

    if plt == "Windows":
        name = 'adb.exe'
    elif plt == "Linux":
        name = 'adb'
    elif plt == "Darwin":
        name = 'adb'

    # 2. use buildin adb
    bin_dir = resource_filename("enhanced_adb", "bin")
    exe = os.path.join(bin_dir, name)
    if os.path.isfile(exe) and _is_valid_exe(exe):
        return exe

    raise RuntimeError("No adb exe could be found. Install adb on your system")


def get_aapt_exe():
    # 1. find in $PATH
    exe = whichcraft.which("aapt")
    if exe and _is_valid_exe(exe):
        return exe

    # 2. use buildin adb
    bin_dir = resource_filename("enhanced_adb", "bin")

    plt = platform.system()

    if plt == "Windows":
        name = 'aapt.exe'
    elif plt == "Linux":
        name = 'aapt'
    elif plt == "Darwin":
        name = 'aapt_darwin'

    exe = os.path.join(bin_dir, name)
    if os.path.isfile(exe) and _is_valid_exe(exe):
        return exe

    raise RuntimeError("No aapt exe could be found. Install aapt on your system")


def _popen_kwargs(prevent_sigint=False):
    startupinfo = None
    preexec_fn = None
    creationflags = 0
    if sys.platform.startswith("win"):
        # Stops executable from flashing on Windows (see imageio/imageio-ffmpeg#22)
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    if prevent_sigint:
        # Prevent propagation of sigint (see imageio/imageio-ffmpeg#4)
        # https://stackoverflow.com/questions/5045771
        if sys.platform.startswith("win"):
            creationflags = 0x00000200
        else:
            preexec_fn = os.setpgrp  # the _pre_exec does not seem to work
    return {
        "startupinfo": startupinfo,
        "creationflags": creationflags,
        "preexec_fn": preexec_fn,
    }


def _is_valid_exe(exe: str):
    cmd = [exe, "version"]
    try:
        with open(os.devnull, "w") as null:
            subprocess.check_call(
                cmd, stdout=null, stderr=subprocess.STDOUT, **_popen_kwargs()
            )
        return True
    except (OSError, ValueError, subprocess.CalledProcessError):
        return False


def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, 'Yi', suffix)



import io
import os

from rich.style import Style
from rich.table import Table
from rich import box
from datetime import datetime

from rich.theme import Theme

from .inspect_result import InspectResult
from .. import _utils


def rich_text(ir: InspectResult):
    from rich.console import Console

    theme = Theme({
        "pr.normal": "",
        "pr.warning": "yellow",
        "pr.danger": "bold red"
    })
    console = Console(file=io.StringIO(), force_terminal=True, theme=theme)

    from rich.panel import Panel
    console.print(Panel("[yellow]{} {}({})".format(ir.label, ir.version, ir.version_code)),
                  justify="center")
    console.line()

    # foundations
    foundation_table = Table.grid(padding=1, pad_edge=True)
    foundation_table.add_column("c0", style="repr.tag_name")
    foundation_table.add_column("c1", style="repr.tag_contents")

    stat = os.stat(ir.path)
    dt = datetime.utcfromtimestamp(stat.st_ctime).strftime('%Y-%m-%d %H:%M:%S')
    foundation_table.add_row("hash", ir.file_hash.upper())
    foundation_table.add_row("package", ir.package)
    foundation_table.add_row("size", _utils.sizeof_fmt(stat.st_size))
    foundation_table.add_row("create_date", str(dt))
    foundation_table.add_row("target_sdk_version", ir.compile_config['targetSdkVersion'])
    foundation_table.add_row("sdk_version", ir.compile_config['sdkVersion'])
    foundation_table.add_row("compile_sdk_version", ir.compile_config['compileSdkVersion'])

    console.print(foundation_table)
    console.line()

    permission_table = Table(show_header=False, show_lines=True, box=box.SIMPLE)
    permission_table.add_column("name", justify="left", overflow="fold")
    permission_table.add_column("description", overflow="fold")

    level_style = {
        '0': "pr.normal",
        '1': "pr.warning",
        '2': "pr.danger"
    }

    count = [0, 0, 0, 0]

    for p in ir.permissions:
        count[0] += 1

        if p.level == '0':
            count[1] += 1
        elif p.level == '1':
            count[2] += 1
        elif p.level == '2':
            count[3] += 1

        style = level_style.get(p.level, 'pr.normal')
        permission_table.add_row("[{}]{}".format(style, p.name), "[{}]{}".format(style, p.description))

    permission_table.title = "(total)/(normal)/(warning)/(danger): {}/{}/{}/{}".format(*count)

    console.print(permission_table)

    console.line()

    from .manifest_parser import ManifestWrapper
    mw = ManifestWrapper(ir.manifest)
    app = mw.application

    activities = app.activities

    for a in activities:
        if a.is_launch_activity:
            console.print("(ACTIVITY) '{}' #Launch".format(a.name))
        else:
            console.print("(ACTIVITY) '{}'".format(a.name))

    console.line()

    services = app.services

    for s in services:
        console.print("(SERVICE) '{}'".format(s.name))

    receivers = app.receivers

    console.line()
    for r in receivers:
        console.print("(RECEIVER) '{}'".format(r.name))

    providers = app.providers

    console.line()
    for p in providers:
        console.print("Provider '{}'".format(p.name))

    metas = app.metas

    console.line()
    for m in metas:
        console.print("Meta '{}'={}".format(m.name, m.value))

    # console.print(ir.properties)

    # console.rule("summary")
    print(console.file.getvalue())  # type: ignore



import logging
import subprocess

from .._utils import get_aapt_exe


class AAPT:

    @staticmethod
    def cmd(args=None):
        logging.debug([get_aapt_exe()] + args)
        return subprocess.run([get_aapt_exe()] + args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)



import os


class InspectResult:

    def __init__(self, path):
        self._path = path
        self._package = ''
        self._label = ''
        self._version = ''
        self._version_code = ''
        self._permissions = []
        self._compile_config = {}
        self._properties = {}
        self._manifest = None

        import hashlib
        block_size = 65536
        hasher = hashlib.sha1()
        with open(path, 'rb') as file:
            buf = file.read(block_size)
            while len(buf) > 0:
                hasher.update(buf)
                buf = file.read(block_size)
        self._file_hash = hasher.hexdigest()

    @property
    def packpage(self):
        return self._package

    @packpage.setter
    def package(self, value):
        self._package = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = value

    @property
    def version_code(self):
        return self._version_code

    @version_code.setter
    def version_code(self, value):
        self._version_code = value

    def add_permission(self, permission):
        self._permissions.append(permission)

    def set_compile_setting(self, compile, sdk, target):
        self._compile_config = {'compileSdkVersion': compile, 'sdkVersion': sdk, 'targetSdkVersion': target}

    @property
    def compile_config(self):
        return self._compile_config

    @property
    def path(self):
        return self._path

    @property
    def permissions(self):
        return self._permissions

    @property
    def file_hash(self):
        return self._file_hash

    def add_properties(self, filename, text):
        self._properties[filename] = text

    @property
    def properties(self):
        return self._properties

    @property
    def manifest(self):
        return self.manifest

    @manifest.getter
    def manifest(self):
        return self._manifest

    @manifest.setter
    def manifest(self, value):
        self._manifest = value



import argparse
import logging
import sys
from logging import basicConfig, CRITICAL, ERROR, WARNING, INFO, DEBUG

from rich.logging import RichHandler


def inspect_apk(args):
    from .inspect.inspecter import do_inspect
    result = do_inspect(args.path)

    if result is None:
        return False

    from .inspect.output import rich_text
    rich_text(result)
    return True


def main(argv=None):

    if argv is None:
        argv = sys.argv[1:]

    from rich.traceback import install
    install()

    parser = argparse.ArgumentParser("eadb")

    parser.add_argument("-v", "--verbose", action="count")
    parser.add_argument("-q", "--quiet", action="count")

    subparsers = parser.add_subparsers(title='support commands')

    # inspect
    inspect = subparsers.add_parser('inspect', help='inspect apk file, usage: eadb inspect path/to/apk')
    inspect.add_argument('path', metavar='APK_PATH', help='the apk path')

    inspect.set_defaults(func=inspect_apk)

    # analyze catcher
    analyze_catcher = subparsers.add_parser('analyze_cather')

    # foo
    foo = subparsers.add_parser('foo')

    args = parser.parse_args(argv)

    log_level = WARNING

    if args.verbose:
        log_level = DEBUG

    if args.quiet:
        log_level = CRITICAL

    FORMAT = "%(message)s"
    basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])

    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_usage()


if __name__ == '__main__':
    main()



import logging
import os
import re
import sys
import yaml

from .inspect_result import InspectResult
from pkg_resources import resource_filename
from .manifest_parser import Manifest

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

'''
regular express

'''
re_packageName = re.compile(r"package: name=\'([^\']+)\'")
re_versionCode = re.compile(r"versionCode=\'([^\']+)\'")
re_versionName = re.compile(r"versionName=\'([^\']+)\'")

re_compileSdkVersion = re.compile(r"compileSdkVersion=\'([^\']+)\'")
re_sdkVersion = re.compile(r"sdkVersion:\'([^\']+)\'")
re_targetSdkVersion = re.compile(r"targetSdkVersion:\'([^\']+)\'")

re_permission_group = re.compile(r"uses-permission:\s([^\n]+)")
re_permission_name = re.compile(r"name=\'([^\']+)\'")

re_label = re.compile(r"application: label=\'([^\']+)\'")

permission_table_file = os.path.join(resource_filename("enhanced_adb", "res"), "permission.yml")
with open(permission_table_file, encoding="utf-8") as file:
    permission_table = yaml.load(file, Loader=Loader)
    permission_data = permission_table['data']


class Permission:
    def __init__(self, name):
        self._name = name

        item = permission_data.get(name)

        if item:
            self._level = item.get('level')
            self._description = item.get('description')
        else:
            self._level = 'Unknown'
            self._description = 'Unknown'

    @property
    def name(self):
        return self._name

    @property
    def level(self):
        return self._level

    @level.getter
    def level(self):
        return str(self._level)

    @property
    def description(self):
        return self._description


def _check_path_and_format(path):
    if os.path.isabs(path):
        apk_path = path
    else:
        apk_path = os.path.abspath(path)

    if not os.path.exists(apk_path):
        logging.error("file \"{0}\" not found".format(apk_path))
        return None

    return apk_path


def _get_re_group_result(re_exp, text):
    try:
        g = re_exp.search(text).groups()
        if not g or len(g) == 0:
            print("can't found {}".format(re_exp), file=sys.stderr)
            raise RuntimeError

        return g[0]
    except AttributeError:
        return None


def do_inspect(arg_path):
    logging.debug("inspect [{}]...".format(arg_path))
    # check path
    logging.debug("check path")
    abs_path = _check_path_and_format(arg_path)
    if not abs_path:
        return None

    result: InspectResult = InspectResult(abs_path)

    import zipfile
    from zipfile import BadZipFile

    apk = None
    try:
        apk = zipfile.ZipFile(abs_path, 'r')
    except BadZipFile:
        logging.error("File is not a apk or zip file")

    if not apk:
        return None

    from ..aapt.aapt import AAPT

    from io import StringIO
    stream = StringIO()
    handler = logging.StreamHandler(stream)
    log = logging.getLogger('root')
    log.addHandler(handler)

    cmd = AAPT.cmd(["dump", "badging", abs_path])

    err = cmd.stderr

    if err:
        logging.error(err.decode("utf-8"))
        return None

    badging_result = cmd.stdout.decode("utf-8")

    label_str = _get_re_group_result(re_label, badging_result)
    logging.debug("label: {}".format(label_str))
    result.label = label_str

    packageName_str = _get_re_group_result(re_packageName, badging_result)
    logging.debug("package_name: {}".format(packageName_str))
    result.package = packageName_str

    versionName_str = _get_re_group_result(re_versionName, badging_result)
    logging.debug("version_name: {}".format(versionName_str))
    result.version = versionName_str

    versionCode_str = _get_re_group_result(re_versionCode, badging_result)
    logging.debug("version_code: {}".format(versionCode_str))
    result.version_code = versionCode_str

    compileSdkVersion_str = _get_re_group_result(re_compileSdkVersion, badging_result)
    targetSdkVersion_str = _get_re_group_result(re_targetSdkVersion, badging_result)
    sdkVersion_str = _get_re_group_result(re_sdkVersion, badging_result)
    result.set_compile_setting(compileSdkVersion_str, sdkVersion_str, targetSdkVersion_str)

    permissions_result = re_permission_group.findall(badging_result)

    for perm in permissions_result:
        name_str = _get_re_group_result(re_permission_name, perm)
        result.add_permission(Permission(name_str))

    manifest_result = AAPT.cmd(["dump", "--values", "xmltree", abs_path, "AndroidManifest.xml"])

    if not manifest_result.stderr:
        manifest = Manifest(manifest_result.stdout.decode("utf-8"))
        result.manifest = manifest

    files = apk.filelist
    for file in files:
        if file.filename.endswith('.properties'):
            text = apk.open(file, "r").read().decode("utf-8")
            result.add_properties(file.filename, text)

    return result



import os

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="enhanced_adb",
    version="0.0.4",
    author="roy.fu",
    author_email="fuzhongqing1995@gmail.com",
    description="enhancement adb",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/royfu/enhanced-adb/src",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        'Development Status :: 3 - Alpha',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': ['eadb = enhanced_adb.command_line:main'],
    },
    python_requires='>=3.6',
    include_package_data=True,
    package_data={
        "enhanced_adb": ["bin/*", "res/*"]
    },
    install_requires=['whichcraft', 'rich', 'pyyaml'],
    py_modules=[],
)
